package com.bankofamerica.crme.clientservice.service.impl;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;

import org.apache.commons.lang.exception.ExceptionUtils;
import org.apache.log4j.Logger;
import org.jbpm.api.ExecutionService;
import org.jbpm.api.IdentityService;
import org.jbpm.api.NewDeployment;
import org.jbpm.api.ProcessInstance;
import org.jbpm.api.ProcessInstanceQuery;
import org.jbpm.api.RepositoryService;
import org.jbpm.api.TaskQuery;
import org.jbpm.api.TaskService;
import org.jbpm.api.identity.Group;
import org.jbpm.api.task.Task;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.io.ClassPathResource;
import org.springframework.orm.hibernate3.HibernateTemplate;
//import org.jbpm.pvm.internal.cfg.SpringConfiguration;
import org.jbpm.pvm.internal.processengine.SpringHelper;


import com.bankofamerica.crme.clientservice.service.ClientService;
import com.bankofamerica.crme.clientservice.service.WorkflowService;

public class WorkflowServiceImpl implements WorkflowService {

	private List<String> processDefinitions;
	private static final Logger logger = Logger.getLogger(WorkflowServiceImpl.class);
	//@Resource(name = "jbpmConfiguration")
	//private SpringConfiguration springConfiguration;
	
	private SpringHelper springHelper;
	
	@Autowired
	private RepositoryService repositoryService;
	
	@Autowired
	private ExecutionService executionService;
	
	@Autowired
	private TaskService taskService;
	
	@Autowired
	private IdentityService idService;

	@Autowired
	private HibernateTemplate hibernateTemplate;	
		
	@PostConstruct
	public void setupProcessDefinitions() {
		try {
			for (String processDefinition : processDefinitions) {
				NewDeployment deployment = repositoryService.createDeployment();
				deployment.addResourceFromUrl(new ClassPathResource(processDefinition)
				.getURL());
				deployment.deploy();
			}
		} catch (IOException e) {
			logger
			.info("IOException occurred: " + ExceptionUtils.getFullStackTrace(e));
			throw new RuntimeException(
					"An error occured while trying to deploy a process definition", e);
		}
	}
	
	/**
	* A simple mutator to set process definitions
	* 
	* @param pd
	*/
	public void setProcessDefinitions(List<String> pd) {
		this.processDefinitions = pd;
	}

	@Override
	public ProcessInstance createNewProcessInstance(String flowId, String businessKey) {

		ProcessInstance pi = executionService.startProcessInstanceByKey(flowId, businessKey);
		return pi;
	}	
	
    /**
     * Returns list of business keys(request numbers) matching the process key and state
     * 
     * @param processKey
     * @param state
     * @return 
     */
	public List<String> listProcessInstanceByKeyAndState(String processKey, String state)
	{
		// create a ProcessInstanceQuery and call processInstanceKey() which will give you list of all 
		// instances. Filter out the instances which are in "ClientAdded" state

		List<String> list = new ArrayList<String>();
		
		ProcessInstanceQuery query = executionService.createProcessInstanceQuery();

		// process id = process name + "-" + version
		// version of deployed process is 1
		List<ProcessInstance> piList = query.processDefinitionId(processKey+"-"+"1").list();
		logger.info("***************** piList found and size is " + piList.size());
		// filter out the processes by state
		for (ProcessInstance pi : piList) {
			if ( pi.findActiveExecutionIn(state)!= null)
			{
				// PI in the wanted state ;-)
				list.add(pi.getKey());
			}
		}		
		return list;
	}
	
    /**
     * Find a process instance/execution and signal it to move     
     * 
     * @param key
     * @return new state of the process instance
     */
	public String signalExecutionByBusinessKey(String key){
		// defaulted the state with error. no, m not being pessimistic :-)
		String newState = "Error - failed to signal the instance";
		
		// query the process instance
		ProcessInstanceQuery query = executionService.createProcessInstanceQuery();
		ProcessInstance pi = query.processInstanceKey(key).list().get(0);
		
		
		if(pi != null){
		// signal PI to move forward
			String id = pi.getId();
			
			if (id == null)	{
				 id = pi.findActiveExecutionIn("ClientAdded").getId();
			}
			pi = executionService.signalExecutionById(pi.getId());
			//newState = pi.geta
			
		}
		return pi.isActive(ClientService.STATE_CLIENT_UPDATED)?ClientService.STATE_CLIENT_UPDATED:newState;
	}

    /**
     * Find all process instances/executions in given state and signal them to move   
     * @param ProcessInstance key
     * @param state 
     */
	public String signalExecutionsFromState(String processKey, String state){
		
		String newState = null;
		ProcessInstanceQuery query = executionService.createProcessInstanceQuery();
		
		// process id = process name + "-" + version
		// version of deployed process is 1
		List<ProcessInstance> piList = query.processDefinitionId(processKey+"-"+"1").list();
		// filter out the processes by state
		for (ProcessInstance pi : piList) {
			if ( pi.isActive(state))
			{
				// signal the right PIs
				pi = executionService.signalExecutionById(pi.getId());
				newState = pi.findActiveActivityNames().iterator().next();
			}
		}
		return newState;
	}
	
	/**
	 * Service returns Pending requests based on search criteria
	 *  
	 *  @param Process key
	 *  @param user group
	 *  
	 * @return Task list with CreateClient request numbers 
	 */
	@Override
	public List<String> getOpenTasksByProcessKeyAndGroup(String processKey, String userGroup){
		
		// TODO
		Group group = idService.findGroupById(userGroup);
	
		// Task service provides an API to query Group tasks 
		taskService.findGroupTasks(userGroup);
		
		return null;
	}
	
	/**
	 * Service creates a user and group and then add user under the given group. 
	 * 
	 * @param group
	 * @param user
	 * 
	 * @return true if operation is successful, false otherwise
	 */
	public boolean addUserInGroup(String group, String user){
		idService.createGroup(group);
		idService.createUser(user, user, user);
		idService.createMembership(user, group);	
		return idService.findGroupsByUser(user).get(0).equals(group);
	}

	/**
	 * Service returns Pending requests based on search criteria
	 *  
	 *  @param activity
	 *  @param user id
	 *  
	 * @return Task list with request numbers
	 */
	@Override
	public List<String> getOpenTasksByActivityAndUser(String activity, String user){
		
		List<String> list = new ArrayList<String>();
		
		List<Task> taskList = taskService.findPersonalTasks(user); 
		// filter out the processes by state
		for (Task t : taskList) {
			if ( t.getActivityName().equals(activity))
			{
			// task associated with the wanted activity ;-)
			// get the CreateClient request number  
			 list.add(executionService.findProcessInstanceById(t.getExecutionId()).getKey());
			}
		}		
		return list;
	}	
	
	/**
	 * Service to facilitate a decision making in CreateClient process. it also completes the task 
	 * assigned to user group.
	 * 
	 * @param userId
	 * @param processKey
	 * @param businessKey
	 * @param decision
	 * 
	 * @return true if operation is successful, false otherwise 
	 */
	
	public boolean takeDecisionForRequest(String userId, String processKey, 
				String businessKey ,String decision){

		boolean exitCode = false;
		
		ProcessInstanceQuery piQuery = executionService.createProcessInstanceQuery();
		ProcessInstance pi = piQuery.processInstanceKey(businessKey).list().get(0);

		// find the Task by process instance id and user id - since the task is assigned to user group
		// user associated with group will also be candidate for the task
		TaskQuery tQuery = taskService.createTaskQuery().processInstanceId(pi.getId()).candidate(userId);
		
		if (tQuery.list().size() > 0)
			{
				// take the task and complete it
				Task task = tQuery.list().get(0);
				taskService.takeTask(task.getId(), userId);
				Map<String,Object> var = new HashMap<String,Object>();
				var.put("decision", decision);
				taskService.completeTask(task.getId(), var);
				exitCode = true;
			}
		return exitCode;
	}
	


}
